import pytest
from airflow_pydantic.airflow import PoolNotFound
from pydantic import ValidationError

from airflow_balancer import BalancerConfiguration, Host, Port
from airflow_balancer.testing import pools


class TestConfig:
    def test_no_duplicate_ports(self):
        with pools(side_effect=PoolNotFound()):
            h1 = Host(name="host1")
            h2 = Host(name="host2")
            p1 = Port(host=h1, port=1002)
            p2 = Port(name="p2", host=h2, port=1000)
            p3 = Port(name="p3", host=h2, port=999)

            with pytest.raises(ValidationError):
                Port()
            with pytest.raises(ValidationError):
                Port(host=Host(name="h1"), host_name="h2")

            p1dupe = Port(host_name="host1", port=1002)

            assert p2 < p3
            assert p2 == p2
            assert p1 == p1dupe
            assert hash(p1) != hash(p2)
            assert p1.pool == "host1-1002"
            assert p2.pool == "p2"
            assert p1dupe.pool == "host1-1002"

            BalancerConfiguration(
                hosts=[h1, h2],
                ports=[p1, p2],
            )

            with pytest.raises(ValueError):
                BalancerConfiguration(
                    hosts=[h1, h2],
                    ports=[p1, p2, p1dupe],
                )



from pathlib import Path

from airflow_config import load_config
from airflow_pydantic.airflow import PoolNotFound

from airflow_balancer import BalancerConfiguration
from airflow_balancer.testing import pools


class TestConfig:
    def test_load_config(self):
        with pools(return_value="Test"):
            config = BalancerConfiguration()
            assert config

        with pools(side_effect=PoolNotFound()):
            config = BalancerConfiguration()
            assert config

    def test_load_config_direct(self):
        with pools():
            fp = Path(__file__).parent.resolve() / "config" / "extensions" / "default.yaml"
            config = BalancerConfiguration.load_path(fp)
            print(config)
            assert config
            assert isinstance(config, BalancerConfiguration)
            assert len(config.hosts) == 4

    def test_load_config_direct_via_airflow_config(self):
        with pools():
            fp = Path(__file__).parent.resolve() / "config" / "extensions" / "default.yaml"
            config = BalancerConfiguration.load(config_name=fp.stem, config_dir=fp.parent)
            print(config)
            assert config
            assert isinstance(config, BalancerConfiguration)
            assert len(config.hosts) == 4

    def test_load_config_direct_via_airflow_config_fallback(self):
        with pools():
            fp = Path(__file__).parent.resolve() / "config" / "extensions" / "default.yaml"
            config = BalancerConfiguration.load(config_name=fp, config_dir="")
            print(config)
            assert config
            assert isinstance(config, BalancerConfiguration)
            assert len(config.hosts) == 4

    def test_load_config_serialize(self):
        # Test serialization needed by the viewer
        with pools():
            fp = str(Path(__file__).parent.resolve() / "config" / "extensions" / "balancer.yaml")
            config = BalancerConfiguration.load_path(fp)
            assert config
            assert isinstance(config, BalancerConfiguration)
            assert len(config.hosts) == 4
            config.model_dump_json(serialize_as_any=True)

    def test_load_config_hydra(self):
        with pools():
            config = load_config("config", "config")
            assert config
            assert "balancer" in config.extensions
            assert len(config.extensions["balancer"].hosts) == 4
            assert [x.name for x in config.extensions["balancer"].hosts] == ["host0", "host1", "host2", "host3"]
            assert config.extensions["balancer"].default_username == "test"
            for host in config.extensions["balancer"].hosts:
                assert host.hook()



from contextlib import contextmanager
from unittest.mock import MagicMock, patch

__all__ = ("pools", "variables")


@contextmanager
def pools(return_value=None, side_effect=None):
    with (
        patch("airflow_pydantic.airflow.Pool.get_pool") as get_pool_mock,
        patch("airflow_pydantic.airflow.Pool.create_or_update_pool") as create_or_update_pool_mock,
        patch("airflow_pydantic.airflow.get_parsing_context") as context_mock,
    ):
        get_pool_mock.return_value = return_value
        create_or_update_pool_mock.return_value = return_value
        if side_effect:
            get_pool_mock.side_effect = side_effect
        context_mock.return_value = MagicMock()
        context_mock.return_value.dag_id = "airflow_balancer.testing.pools"
        yield get_pool_mock


@contextmanager
def variables(return_value=None, side_effect=None):
    with patch("airflow_pydantic.airflow.Variable.get") as get_mock:
        get_mock.return_value = return_value
        if side_effect:
            get_mock.side_effect = side_effect
        yield get_mock



from __future__ import annotations

import os

from airflow.configuration import conf
from airflow.plugins_manager import AirflowPlugin
from airflow.security import permissions
from airflow.www.auth import has_access
from flask import Blueprint, request
from flask_appbuilder import BaseView, expose

from .functions import get_hosts_from_yaml, get_yaml_files

__all__ = (
    "AirflowBalancerViewerPluginView",
    "AirflowBalancerViewerPlugin",
)


class AirflowBalancerViewerPluginView(BaseView):
    """Creating a Flask-AppBuilder View"""

    default_view = "home"

    @expose("/hosts")
    @has_access([(permissions.ACTION_CAN_READ, permissions.RESOURCE_WEBSITE)])
    def hosts(self):
        """Create hosts view"""
        yaml = request.args.get("yaml")
        if not yaml:
            return self.render_template("airflow_config/500.html", yaml="- yaml file not specified")
        try:
            config = get_hosts_from_yaml(yaml)
        except FileNotFoundError:
            return self.render_template("airflow_balancer/500.html", yaml=yaml)
        return self.render_template("airflow_balancer/hosts.html", config=config)

    @expose("/")
    @has_access([(permissions.ACTION_CAN_READ, permissions.RESOURCE_WEBSITE)])
    def home(self):
        """Create default view"""
        # Locate the dags folder
        dags_folder = os.environ.get("AIRFLOW__CORE__DAGS_FOLDER", conf.getsection("core").get("dags_folder"))
        if not dags_folder:
            return self.render_template("airflow_balancer/404.html")
        yamls, yamls_airflow_config = get_yaml_files(dags_folder)
        return self.render_template("airflow_balancer/home.html", yamls=yamls, yamls_airflow_config=yamls_airflow_config)


# Instantiate a view
airflow_balancer_viewer_plugin_view = AirflowBalancerViewerPluginView()

# Creating a flask blueprint
bp = Blueprint(
    "Airflow Balancer",
    __name__,
    template_folder="templates",
    static_folder="static",
    static_url_path="/static/airflow-balancer",
)

# Create menu items
docs_link_subitem = {
    "label": "Airflow Balancer Docs",
    "name": "Airflow Balancer Docs",
    "href": "https://airflow-laminar.github.io/airflow-balancer/",
    "category": "Docs",
}

view_subitem = {"label": "Airflow Balancer Viewer", "category": "Laminar", "name": "Laminar", "view": airflow_balancer_viewer_plugin_view}


class AirflowBalancerViewerPlugin(AirflowPlugin):
    """Defining the plugin class"""

    name = "Airflow Balancer"
    flask_blueprints = [bp]
    appbuilder_views = [view_subitem]
    appbuilder_menu_items = [docs_link_subitem]



import os
from pathlib import Path

from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates

# from airflow import __file__ as airflow_root
from flask_appbuilder import __file__ as flask_appbuilder_root
from uvicorn import run

from ..functions import get_hosts_from_yaml, get_yaml_files

_static_path = Path(__file__).parent.parent / "static"
_static_airflow_path = Path(__file__).parent / "static"

__all__ = ("build_app", "main")


class AD(dict):
    __slots__ = ()
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__


def _url_for(name: str, filename: str, **kwargs) -> str:
    if name == "Airflow Balancer.static":
        return f"/static/{filename}"
    return ""


def build_app() -> FastAPI:
    # Create a FastAPI instance
    app = FastAPI()

    # Serve static files
    app.mount(
        "/static",
        StaticFiles(directory=_static_path),
        name="static",
    )

    app.mount(
        "/static_airflow",
        StaticFiles(directory=_static_airflow_path),
        name="static_airflow",
    )

    # Templates
    templates = Jinja2Templates(
        directory=[
            Path(__file__).parent.parent / "templates" / "airflow_balancer",
            Path(__file__).parent / "templates",
            Path(flask_appbuilder_root).parent / "templates",
        ],
        autoescape=True,
    )

    # Mount top level routes
    # @app.get("/favicon.ico", include_in_schema=False, response_class=FileResponse)
    # async def readFavicon():
    #     return FileResponse(_static_path / "img" / "favicon.png")

    # Facsimile of Flask Appbuilder + Airflow templates
    fab_common_mock = {
        "request": AD(url_for=_url_for),
        "base_template": "airflow_mock.html",
        "appbuilder": AD(
            app_theme="",
            app_name="airflow-balancer",
            app_icon="",
            menu=AD(get_list=lambda: [], extra_classes=None),
            languages={},
            get_url_for_index="/",
            get_url_for_login="",
        ),
        "get_flashed_messages": lambda *args, **kwargs: [],
        "current_user": AD(is_anonymous=True),
        "session": {"locale": "en"},
        "_": lambda *args, **kwargs: "",
    }

    @app.get("/")
    async def home():
        dags_folder = os.environ.get("AIRFLOW__CORE__DAGS_FOLDER", os.environ.get("AIRFLOW_HOME", Path(__file__).parent.parent.parent / "tests"))
        if not dags_folder:
            return templates.TemplateResponse("airflow_balancer/404.html", fab_common_mock)
        yamls, yamls_airflow_config = get_yaml_files(dags_folder)
        return templates.TemplateResponse("home.html", {"yamls": yamls, "yamls_airflow_config": yamls_airflow_config, **fab_common_mock})

    @app.get("/hosts")
    async def hosts(yaml: str = ""):
        if not yaml:
            return templates.TemplateResponse("airflow_balancer/500.html", {"yaml": "- yaml file not specified", **fab_common_mock})
        config = get_hosts_from_yaml(yaml)
        return templates.TemplateResponse("hosts.html", {"config": config, **fab_common_mock})

    return app


def main():
    app = build_app()

    # Run the application using Uvicorn
    run(app, host="0.0.0.0", port=int(os.environ.get("PORT", 8000)))
